"""
"""
# Future Imports
from __future__ import annotations

# Local
from ..satlib import Source, Posiform
from ..symbols import T_NOT, T_AND, T_OR, T_XOR, T_IMP, T_IFF, T_RIMP, T_EQ, T_NE, T_LT, T_LE, T_GT, T_GE, T_IDX, T_ADD, T_SUB, T_DIV, T_MUL, T_NEG


class MetaSatType(type):

    base_type = None

    def __new__(cls, name: str, bases: tuple, namespace: dict):
        if cls.base_type is None:
            if name == "SatType":
                cls.base_type = type(name, bases, namespace)
                return cls.base_type
            else:
                raise NotImplementedError(f"'SatType must be implemented before '{name}'")
        else:
            if name == "Number":
                cls.base_type.Number = type(name, bases, namespace)
                return cls.base_type.Number
            elif name == "Expr":
                cls.base_type.Expr = type(name, bases, namespace)
                # -*- Some Magic -*- A-bra-ca-da-bra -*-
                cls.base_type.Expr.RULES.update(cls.base_type.Expr.get_rules())
                return cls.base_type.Expr
            else:
                return type(name, bases, namespace)


class SatType(metaclass=MetaSatType):
    """"""

    def __init__(self, *, source: Source = None, lexpos: int = None):
        if isinstance(source, Source):
            source.track(self, lexpos)
        elif source is None:
            Source.blank(self)
        else:
            raise TypeError(f"Invalid type '{type(source)}' for 'source' argument")

    @property
    def is_number(self) -> bool:
        return False

    @property
    def is_array(self) -> bool:
        return False

    @property
    def is_expr(self) -> bool:
        return False

    @property
    def is_var(self) -> bool:
        return False

    # -*- Operator Definition -*-

    # :: Aliases
    def __invert__(self):
        return self._NOT_()

    def __and__(self, other):
        return self._AND_(other)

    def __or__(self, other):
        return self._OR_(other)

    # :: Logical
    def _NOT_(self) -> SatType:
        return self.__class__.Expr(T_NOT, self)

    def _AND_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_AND, self, other)

    def _OR_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_OR, self, other)

    def _XOR_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_XOR, self, other)

    def _IMP_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_IMP, self, other)

    def _IFF_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_IFF, self, other)

    def _RIMP_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_RIMP, self, other)

    # :: Comparison
    def _EQ_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_EQ, self, other)

    def _NE_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_NE, self, other)

    def _LT_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_LT, self, other)

    def _LE_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_LE, self, other)

    def _GT_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_GT, self, other)

    def _GE_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_GE, self, other)

    # :: Indexing
    def _IDX_(self, i: tuple) -> SatType:
        return self.__class__.Expr(T_IDX, self, *i)

    # :: Arithmetic

    def _ADD_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_ADD, self, other)

    def _SUB_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_SUB, self, other)

    def _DIV_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_DIV, self, other)

    def _MUL_(self, other: SatType) -> SatType:
        return self.__class__.Expr(T_MUL, self, other)

    def _NEG_(self) -> SatType:
        return self.__class__.Expr(T_NEG, self)


__all__ = ["SatType"]



from satyrus import SatAPI, Posiform

class csv(SatAPI):
    """"""

    ext: str = "csv"

    def solve(self, posiform: Posiform, **params: dict) -> str:
        """"""
        lines = []

        for term, cons in posiform:
            if term is None:
                lines.append(f"{cons:.5f}")
            else:
                lines.append(",".join([f"{cons:.5f}", *term]))

        return "\n".join(lines)


from .main import arange, log, prompt
from .package import package_path
from .performance import Timing
from .posiform import Posiform
from .pythonshell import PythonShell, PythonError
from .source import Source
from .stack import Queue, Stack

__all__ = [
    "arange",
    "log",
    "prompt",
    "package_path",
    "Timing",
    "Posiform",
    "PythonShell",
    "PythonError",
    "Source",
    "Queue",
    "Stack",
]



"""
satyrus/compiler/memory.py
--------------------------
"""

## Local
from ..error import SatReferenceError
from ..types import Var, SatType

class Memory(object):
	""" Satyrus Compiler Memory
	"""
	def __init__(self, defaults: dict=None):
		self.__memory = [{} if defaults is None else defaults]

	def __str__(self):
		return ">\n".join("\n".join(f"{key}:\t{val}" for key, val in layer.items()) for layer in self.__memory)

	def __iter__(self):
		i = len(self.__memory) - 1
		found = set()
		while i >= 0:
			for key in self.__memory[i]:
				if key not in found:
					yield self.__memory[i][key]
					found.add(key)
				else:
					continue
			else:
				i -= 1

	def __getitem__(self, key):
		i = len(self.__memory) - 1
		while i >= 0:
			try:
				return self.__memory[i][key]
			except KeyError:
				i -= 1
		else:
			return None

	def __contains__(self, item):
		return self[item] is not None

	@property
	def copy(self):
		new_mem = self.__class__()
		new_mem.__memory = [layer.copy() for layer in self.__memory]
		return new_mem

	def push(self):
		"""
		"""
		self.__memory.append({})

	def pop(self, n: int=1):
		"""
		"""
		for _ in range(n):
			if len(self.__memory) > 1:
				self.__memory.pop()
			else:
				raise ValueError("Can't remove global scope.")

	def clear(self):
		"""
		"""
		# Clear list
		del self.__memory[:]
		
		# Add global scope
		self.__memory.append({})

	def memset(self, name: Var, value: SatType):
		"""
		"""
		self.__memory[-1][name] = value

	def memget(self, name: Var):
		"""
		"""
		value = self[name]
		if value is None:
			raise SatReferenceError(f"Undefined variable `{name}`.", target=name)
		else:
			return value


# Future Imports
from __future__ import annotations

# Standard Library
import json
import importlib.resources as resources
from pathlib import Path

# Third-Party
from cstream import stderr

# Local
from ..error import SatRuntimeError, EXIT_SUCCESS, EXIT_FAILURE
from ..satlib import Source, Posiform, package_path
from ..compiler import SatCompiler
from ..compiler.instructions import INSTRUCTIONS
from ..parser import SatParser
from ..parser.legacy import SatLegacyParser


class Satyrus(object):
    """
    Compiler flow diagram
    ---------------------

    ┌───────────┐
    │source.sat ├──┐
    └───────────┘  │
                   │
          .        │    ┌────────┐
          .        ├────►Satyrus3│      ┌───────────┐
          .        ├────►(Python)├──────►energy.json│
                   │    └────────┘      └───────────┘
    ┌───────────┐  │
    │energy.json├──┘
    └───────────┘
    """

    __cache_path__ = None

    def __init__(self, legacy: bool = False):
        """
        Parameters
        ----------
        legacy : bool (optional)
            If true, opts for the legacy parser and its syntax.
        """

        # Choose parser
        if legacy:
            self.parser = SatLegacyParser()
        else:
            self.parser = SatParser()

        # Start Compiler Engine
        self.compiler = SatCompiler(INSTRUCTIONS, self.parser)

        # Compilation Cache
        self.__cache__ = {}

    @staticmethod
    def _path(path: Path) -> str:
        """"""
        return str(path.absolute())

    @staticmethod
    def _stat(path: Path) -> float:
        """"""
        return path.stat().st_mtime

    @staticmethod
    def _stamp(path: Path) -> tuple[str, float]:
        """"""
        return (Satyrus._path(path), Satyrus._stat(path))

    def compile(self, *paths: str | Path) -> Posiform | None:
        """"""
        if not all(isinstance(path, (str, Path)) for path in paths):
            raise TypeError("File paths must be of type 'str' or 'pathlib.Path'")
        else:
            return self.__compile(*paths)

    def __compile(self, *paths: str) -> Posiform | None:
        """"""
        if not paths:
            raise IOError("No input files provided")
        else:
            # Status Code
            code = EXIT_SUCCESS

            # Load Cache
            self.load_cache()

            # Empty Energy Equation
            source_list: list[Path] = []
            energy_list: list[Path] = []

            for path in paths:
                path = Path(path)
                if not path.exists() or not path.is_file():
                    stderr[0] << f"Error: File '{path}' does not exists"
                    code |= EXIT_FAILURE
                elif path.suffix == ".sat":
                    source_list.append(path)
                elif path.suffix == ".json":
                    energy_list.append(path)
                else:
                    stderr[0] << f"Error: Invalid file extension '{path.suffix}'"
                    code |= EXIT_FAILURE

            total_energy = Posiform(None)

            for energy_path in energy_list:
                if not self.cached(energy_path):
                    with energy_path.open(mode="r") as energy_file:
                        try:
                            energy = Posiform.fromJSON(energy_file.read())
                        except json.decoder.JSONDecodeError:
                            stderr[0] << f"Error: Inconsistent JSON data for energy equation in '{energy_path}'"
                            code |= EXIT_FAILURE
                        else:
                            total_energy += self.cache(energy_path, energy)
                else:
                    total_energy += self.cache(energy_path)

            for source_path in source_list:
                if not self.cached(source_path):
                    if self.compiler.compile(Source(fname=source_path)):
                        code |= EXIT_FAILURE
                    else:
                        total_energy += self.cache(source_path, self.compiler.energy)
                else:
                    total_energy += self.cache(source_path)

            self.write_cache()

            if not code:
                return total_energy
            else:
                raise SatRuntimeError("Compilation terminated with errors")

    # -*- Cache -*-
    def cache(self, path: Path, energy: Posiform | None = None) -> Posiform:
        if energy is None:
            return Posiform.fromMiniJSON(self.__cache__[self._path(path)]["energy"])
        else:
            self.__cache__[self._path(path)] = {"stat": self._stat(path), "energy": energy.toMiniJSON()}
            return energy

    def cached(self, path: Path) -> bool:
        key = self._path(path)
        return (key in self.__cache__) and (self.__cache__[key]["stat"] == self._stat(path))

    @classmethod
    def cache_path(cls) -> Path:
        if cls.__cache_path__ is None:
            cls.__cache_path__ = package_path(fname='.satyrus')
        return cls.__cache_path__
        

    def load_cache(self):
        if self.cache_path().exists() and self.cache_path().is_file():
            with self.cache_path().open(mode="r") as file:
                self.__cache__ = json.load(file)
                if self.__cache__ is None:
                    self.__cache__ = {}
        elif not self.cache_path().exists():
            with self.cache_path().open(mode="w") as file:
                json.dump(None, file)
                self.__cache__ = {}
        else:
            raise RuntimeError(f"Unable to load file '{self.cache_path()}'")

    def write_cache(self):
        with self.cache_path().open(mode="w") as file:
            if self.__cache__:
                json.dump(self.__cache__, file)
            else:
                json.dump(None, file)

    @classmethod
    def clear_cache(cls):
        with cls.cache_path().open(mode="w") as file:
            json.dump({}, file)


__all__ = ["Satyrus"]



""" SYS_CONFIG
    ==========

    STATUS: INCOMPLETE
"""

## Standard Library
from pathlib import Path

## Local
from ..compiler import SatCompiler
from ...satlib import Source
from ...symbols import PREC, DIR, LOAD, OUT, EPSILON, ALPHA, EXIT
from ...types import SatType, String, Number, Var, Array
from ...error import SatValueError, SatTypeError, SatFileError, SatWarning


def sys_config(compiler, name: Var, args: list):
    if name in sys_config_options:
        return sys_config_options[name](compiler, name, len(args), args)
    else:
        compiler << SatValueError(f"Invalid config option '{name}'.", target=name)
    compiler.checkpoint()


def sys_config_prec(compiler: SatCompiler, name: Var, argc: int, argv: list):
    if argc != 1:
        if argc == 0:
            compiler << SatValueError(
                f"'?prec' expected 1 argument, got {argc}", target=name
            )
        else:
            compiler << SatValueError(
                f"'?prec' expected 1 argument, got {argc}", target=argv[1]
            )
    elif type(argv[0]) is not Number:
        compiler << SatTypeError(
            f"Precision must be a positive integer.", target=argv[0]
        )
    elif not argv[0].is_int or int(argv[0]) <= 0:
        compiler << SatValueError(
            f"Precision must be a positive integer.", target=argv[0]
        )
    else:
        compiler.env[PREC] = argv[0]
    compiler.checkpoint()


def sys_config_epsilon(compiler: SatCompiler, name: Var, argc: int, argv: list):
    if argc != 1:
        if argc == 0:
            compiler << SatValueError(
                f"'?epsilon' expected 1 argument, got none.", target=name
            )
        else:
            compiler << SatValueError(
                f"'?epsilon' expected 1 argument, got {argc}.", target=argv[1]
            )
    elif not isinstance(argv[0], Number):
        compiler << SatTypeError(f"Epsilon must be a positive number", target=argv[0])
    elif float(argv[0]) <= 0.0:
        compiler << SatValueError(f"Epsilon must be a positive number", target=argv[0])
    else:
        if float(argv[0]) >= 1.0:
            compiler < SatWarning("Large values for Epsilon may cause Integrity errors")
        compiler.env[EPSILON] = argv[0]
    compiler.checkpoint()


def sys_config_load(compiler: SatCompiler, name: Var, argc: int, argv: list):
    if argc == 0:
        compiler << SatValueError(
            "`?load` expected 1 or more arguments, got none.", target=name
        )
    else:
        paths = []
        for fname in argv:
            path = Path(f"{fname}.sat")
            if not path.exists() or not path.is_file():
                compiler << SatFileError(f"file '{path}' not found.", target=fname)
            else:
                paths.append(path.absolute())

        bytecode = []
        for path in paths:
            bytecode.extend(compiler.parse(Source(path)))
        else:
            compiler.execute(bytecode)

        compiler.checkpoint()


def sys_config_alpha(compiler, name: Var, argc: int, argv: list):
    if argc != 1:
        if argc == 0:
            compiler << SatValueError(
                f"`?alpha` expected 1 argument, got none.", target=name
            )
        else:
            compiler << SatValueError(
                f"`?alpha` expected 1 argument, got {argc}.", target=argv[1]
            )
    elif type(argv[0]) is not Number:
        compiler << SatTypeError(
            f"Parameter `alpha` must be a positive number.", target=argv[0]
        )
    elif float(argv[0]) <= 0:
        compiler << SatValueError(
            f"Parameter `alpha` must be a positive number.", target=argv[0]
        )
    else:
        compiler.env[ALPHA] = argv[0]
    compiler.checkpoint()


def sys_config_exit(compiler, name: Var, argc: int, argv: list):
    """Exits program, for debug purposes."""
    if argc != 1:
        if argc == 0:
            compiler << SatValueError(
                f"`?exit` expected 1 argument (exit code), got none.", target=name
            )
        else:
            compiler << SatValueError(
                f"`?exit` expected 1 argument (exit code), got {argc}.", target=argv[1]
            )
    elif type(argv[0]) is not Number:
        compiler << SatTypeError(
            f"The exit code must be a non-negative integer.", target=argv[0]
        )
    elif not argv[0].is_int or int(argv[0]) < 0:
        compiler << SatValueError(
            f"The exit code must be a non-negative integer.", target=argv[0]
        )
    else:
        compiler.exit(int(argv[0]))
    compiler.checkpoint()


# def sys_config_out(compiler, name: Var, argc: int, argv: list):
# raise NotImplementedError

sys_config_options = {
    EXIT: sys_config_exit,
    PREC: sys_config_prec,
    LOAD: sys_config_load,
    EPSILON: sys_config_epsilon,
    ALPHA: sys_config_alpha,
    ## OUT : sys_config_out,
}



"""
"""
# Standard Library
import itertools as it

# Third-Party
from tabulate import tabulate
from cstream import stdlog

# Local
from ..compiler import SatCompiler
from ...error import SatValueError, SatCompilerError, SatWarning
from ...types import Number
from ...symbols import CONS_INT, CONS_OPT, PREC, EPSILON, ALPHA


def run_script(compiler: SatCompiler, *args: tuple):
    """
    """
    # Setup compiler environment
    run_script_setup(compiler)

    # Retrieve constraints
    run_script_constraints(compiler)

    # Check for Errors
    compiler.checkpoint()

    # Compute penalties
    run_script_penalties(compiler)

    # Generate Energy Equations
    run_script_energy(compiler)


def run_script_setup(compiler: SatCompiler):
    ## Set numeric precision
    Number.prec(int(compiler.env[PREC]))

    # value for alpha
    compiler.env[ALPHA] = Number(compiler.env[ALPHA])

    # value for epsilon
    compiler.env[EPSILON] = Number(compiler.env[EPSILON])

    ## Parameter validation
    if compiler.env[EPSILON] < pow(10, -Number.prec(None)):
        compiler << SatValueError(
            "Tiebraker 'epsilon' was neglected due to numeric precision. Choose a greater value",
            target=compiler.env[EPSILON],
        )


def run_script_constraints(compiler: SatCompiler):
    """ """

    # -*- Constraint Validation -*-
    if len(compiler.constraints[CONS_INT]) + len(compiler.constraints[CONS_OPT]) == 0:
        compiler << SatCompilerError("No problem defined. Maybe you are just fine :)", target=compiler.source.eof)
    elif len(compiler.constraints[CONS_INT]) == 0:
        compiler < SatWarning("No Integrity condition defined.", target=compiler.source.eof)
    elif len(compiler.constraints[CONS_OPT]) == 0:
        compiler < SatCompilerError("No Optmization condition defined.", target=compiler.source.eof)


def run_script_penalties(compiler: SatCompiler):
    """ """
    # -*- Penalty Computation -*-
    compiler.penalties.update({0: float(compiler.env[ALPHA])})

    levels: dict[int, float] = {}

    constraints = it.chain(compiler.constraints[CONS_OPT], compiler.constraints[CONS_INT])

    for level, energy in constraints:
        if level not in levels:
            levels[level] = 0.0

        for term, cons in energy:
            if term is None:
                continue
            else:
                # Compute Energy Gap
                levels[level] += abs(cons)
    else:
        levels: list = sorted(levels.items())

    epsilon = float(compiler.env[EPSILON])

    # Compute penalty levels
    level_j, n_j = levels[0]
    for i in range(1, len(levels)):
        # level_k <- level_j, n_k <- n_j, level_j <- level_i, n_j <- n_i
        (level_k, n_k), (level_j, n_j) = (level_j, n_j), levels[i]

        if i == 1: # Base Penalty
            compiler.penalties[level_j] = compiler.penalties[level_k] * n_k + epsilon
        else:
            compiler.penalties[level_j] = compiler.penalties[level_k] * (n_k + 1)

    # -*- Penalty Table Exhibition -*-
    if stdlog[2]:
        stdlog[2] << "PENALTY TABLE"
        stdlog[2] << tabulate(
            [(f"{k}", f"{n}", compiler.penalties[k]) for k, n in levels],
            headers=["lvl", "n", "value"],
            tablefmt="pretty",
        )
        stdlog[2] << ""
        stdlog[2] << "CONSTANTS"
        stdlog[2] << tabulate(
            [[compiler.env[EPSILON], compiler.env[ALPHA]]],
            headers=["ε", "α"],
            tablefmt="pretty",
        )


def run_script_energy(compiler: SatCompiler):
    """"""
    # Integrity
    Ei = sum((compiler.penalties[level] * energy for level, energy in compiler.constraints[CONS_INT]), 0.0)

    # Optimality
    Eo = sum((compiler.penalties[level] * energy for level, energy in compiler.constraints[CONS_OPT]), 0.0)

    compiler.energy = Ei + Eo


# Standard Library
import os

# Third-Party
from cstream import stderr, stdout

# Local
from ..satlib import Source

# Constants
EXIT_SUCCESS = 0
EXIT_FAILURE = 1


class SatError(Exception):
    "Error"

    def __init__(self, msg=None, target=None, code=EXIT_FAILURE):
        Exception.__init__(self, msg)
        self.msg = str(msg) if msg is not None else ""
        self.code = code
        self.target = target

    def __str__(self) -> str:
        if Source.trackable(self.target):
            if self.target.source is None:
                return f"{self.__class__.__doc__}: {self.msg}\n"
            else:
                return self.target.source.error(self.msg, target=self.target, name=self.__doc__)
        else:
            return self.msg


class SatCompilerError(SatError):
    "Compiler Error"


class SatExprError(SatError):
    "Expression Error"


class SatIndexError(SatError):
    "Index Error"


class SatFileError(SatError):
    "File Error"


class SatLexerError(SatError):
    "Lexer Error"


class SatParserError(SatError):
    "Parser Error"


class SatPythonError(SatError):
    "Python Error"


class SatReferenceError(SatError):
    "Reference Error"

class SatRuntimeError(SatError):
    "Runtime Error"

class SatSyntaxError(SatError):
    "Syntax Error"


class SatTypeError(SatError):
    "Type Error"


class SatValueError(SatError):
    "Value Error"


class SatWarning(SatError):
    "Warning"


class SatSolverError(SatError):
    "Solver Error"


class SatExit(SatError):
    "Exit"

    def __init__(self, code: int):
        SatError.__init__(self, f"exit code {code}")


__all__ = [
    "EXIT_SUCCESS",
    "EXIT_FAILURE",
    "SatError",
    "SatCompilerError",
    "SatExprError",
    "SatIndexError",
    "SatFileError",
    "SatLexerError",
    "SatParserError",
    "SatPythonError",
    "SatReferenceError",
    "SatRuntimeError",
    "SatSyntaxError",
    "SatTypeError",
    "SatValueError",
    "SatWarning",
    "SatSolverError",
    "SatExit",
]



class PythonError(Exception):
    ...

class PythonShell:

    def __init__(self):
        self.globals = {**globals()}

    def __call__(self, code: str, evaluate=False):
        if evaluate:
            try:
                bytecode = compile(code, '<pyshell>', 'eval')
                return eval(bytecode, globals=self.globals)
            except Exception as error:
                raise PythonError(*error.args)
        else:
            try:
                bytecode = compile(code, '<pyshell>', 'exec')
                return exec(bytecode, globals=self.globals)
            except Exception as error:
                raise PythonError(*error.args)




"""
/satyrus/api/api.py
"""
# Future Imports
from __future__ import annotations

# Standard Library
import sys
import json
import shutil
import marshal
import traceback
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import final

# Third-Party
from cstream import stdwar, stderr, stdlog

# Local
from ..error import SatRuntimeError, EXIT_FAILURE, EXIT_SUCCESS
from ..satlib import Timing, Posiform, package_path, prompt
from ..satyrus import Satyrus


class SatFinder:
    """"""

    key: str = None
    require: dict = {}

    @classmethod
    def find_spec(cls, name: str, *__args, **__kwargs):
        if cls.key is not None:
            stdwar[1] << f"Warning: missing module {name!r} for {cls.key!r}"
            if cls.key in cls.require:
                cls.require[cls.key].append(name)
            else:
                cls.require[cls.key] = [name]
        return None

    @classmethod
    def set_key(cls, key: str):
        cls.key = key


# -*- Register SatFinder -*- #
sys.meta_path.append(SatFinder)


class MetaSatAPI(ABCMeta):
    """"""

    base_class: MetaSatAPI = None
    __stack__: list[MetaSatAPI] = []
    __satapi__: dict[str, MetaSatAPI] = {}

    def __new__(cls, name: str, bases: tuple, namespace: dict):
        """"""
        if name == "SatAPI":
            if cls.base_class is None:
                cls.base_class = type.__new__(
                    cls,
                    name,
                    bases,
                    {
                        **namespace,
                        "__stack__": cls.__stack__,
                        "__satapi__": cls.__satapi__,
                    },
                )
                return cls.base_class
            else:
                raise NameError("'SatAPI' base class is already implemented")
        elif cls.base_class is None:
            raise NameError("'SatAPI' base class is not implemented yet")
        elif not cls.implements_solve(namespace):
            cls.base_class.error("'solve' method is not implemented")


        interface: MetaSatAPI = type.__new__(cls, name, bases, namespace)

        if name in cls.__satapi__:
            stdwar[1] << f"API Interface '{name}' redefinition"
            cls.__stack__.append(name)
        elif name == "default":
            pass
        else:
            cls.__stack__.append(name)

        cls.__satapi__[name] = interface

        return cls.__satapi__[name]

    @classmethod
    def implements_solve(cls, namespace: dict):
        if 'solve' not in namespace:
            return False
        elif not callable(namespace['solve']):
            return False
        else:
            return True


class SatAPI(metaclass=MetaSatAPI):
    """"""

    __stack__: list = []
    __satapi__: dict = {}
    __satref__: dict = {}
    __loaded__: bool = False
    __energy__: Posiform = None

    ext: str = "out.txt"

    def __init__(self, *paths, guess: dict = None, legacy: bool = False):
        """"""
        if paths:
            try:
                energy: Posiform | None = Satyrus(legacy=legacy).compile(*paths)
            except SatRuntimeError as exc:
                stderr[0] << exc
                energy: Posiform | None = None
        else:
            energy: Posiform | None = None

        if (energy is not None) and (guess is not None):
            energy = energy(guess)

        self.energy(energy)

    @classmethod
    def energy(cls, __energy: Posiform | None):
        cls.__energy__ = __energy

    @classmethod
    def extension(cls, name: str) -> str:
        if name in cls.__satapi__:
            return cls.__satapi__[name].ext
        else:
            return cls.ext

    def __getitem__(self, name: str):
        """"""

        self._load()

        if name in self.__satapi__:
            return self.__satapi__[name]

        if name not in self.__satref__:
            raise KeyError(f"Undefined API interface {name!r}")

        pack_path: Path = package_path()
        code_path: Path = pack_path.joinpath(self.__satref__[name])

        with code_path.open(mode="rb") as file:
            code = marshal.load(file)

        SatFinder.set_key(name)

        exec(code, {})

        SatFinder.set_key(None)

        if name not in self.__satapi__:
            raise KeyError("Defined API not in '__satapi__'")
        elif self.__satapi__[name] is None:
            modules: str = ", ".join(SatFinder.require[name])
            self.error(
                f"Interface {name!r} is unavailable due to missing modules: {modules}"
            )
        else:
            return self.__satapi__[name]

    @Timing.timer(level=2, section="Solver.solve")
    def __call__(self, **params) -> tuple[dict, float] | object:
        """"""

        if "$solver" not in params:
            raise KeyError("No solver specified")

        name: str = params["$solver"]

        if self.__energy__ is None:
            return None
        else:
            return self[name]().solve(self.__energy__, **params)

    @classmethod
    def __load__(cls):
        """"""

        data_path: Path = package_path(fname=".sat-api")

        if not data_path.exists() or data_path.is_dir():
            raise IOError("API metadata not found. Try reinstalling Satyrus")

        with data_path.open(mode="r", encoding="utf8") as file:
            data: dict = json.load(file)

        if data is None:
            cls._bootstrap()
        elif not isinstance(data, dict):
            raise IOError("API metadata is corrupted. Try reinstalling Satyrus")
        else:
            cls.__satref__.update(data)

        cls.__loaded__ = True

    @classmethod
    def _load(cls):
        if not cls.__loaded__:
            cls.__load__()

    @classmethod
    def _bootstrap(cls):
        """"""
        pack_path: Path = package_path()
        cls.add(*pack_path.glob("*.py"))

    @classmethod
    def add(cls, *paths: str):
        """\
        Includes new SatAPI interfaces given a separate Python file (``.py``)

        Parameters
        ----------
        *paths : tuple[str]
            Paths to Python files.
        """

        pack_path: Path = package_path()
        data_path: Path = pack_path.joinpath(".sat-api")

        with data_path.open(mode="r", encoding="utf8") as file:
            data: dict = json.load(file)

        if data is None:
            pass
        elif not isinstance(data, dict):
            raise IOError("API metadata is corrupted. Try reinstalling Satyrus")
        elif not all(isinstance(k, str) for k in data.keys()):
            raise IOError("API metadata is corrupted. Try reinstalling Satyrus")
        elif not all(isinstance(v, str) for v in data.values()):
            raise IOError("API metadata is corrupted. Try reinstalling Satyrus")
        else:
            cls.__satref__.update(data)

        for file_path in map(Path, paths):
            if not file_path.exists() or not file_path.is_file():
                stderr[0] << f"Error: '{file_path}' does not exists"
                continue
            elif file_path.suffix != ".py":
                (
                    stderr[0]
                    << f"Error: '{file_path.absolute()}' is not a valid Python file"
                )
                continue

            new_path = pack_path.joinpath(file_path.name)

            # Copy file to data folder
            try:
                shutil.copyfile(file_path, new_path)
            except shutil.SameFileError:
                # Everything is gonna be alright
                pass

            # Build
            try:
                answer: tuple[str, str] | None = cls._build(new_path)

                if answer is not None:
                    name, path = answer
                    cls.__satref__[name] = path
            finally:
                # Remove Source
                new_path.unlink()
        
        with data_path.open(mode="w", encoding="utf8") as file:
            json.dump(cls.__satref__, file)

    @classmethod
    def build(cls, *paths: str):
        """"""
        for file_path in map(Path, paths):
            cls._build(file_path)

    @classmethod
    def _build(cls, file_path: Path) -> tuple[str, str] | None:
        """ """
        if not file_path.exists() or not file_path.is_file():
            stderr[0] << f"Error: '{file_path}' does not exists"
            return None
        elif file_path.suffix != ".py":
            stderr[0] << f"Error: '{file_path.absolute()}' is not a valid Python file"
            return None

        with file_path.open(mode="r") as file:
            source: str = file.read()

        # Compile Interface definition
        code = compile(source, filename=file_path.name, mode="exec")

        try:
            exec(code, {})

            new: int = len(cls.__stack__)

            if new == 0:
                cls.warn(f"No new interfaces declared in '{file_path.name}'")
                return None
            elif new == 1:
                name: str = cls.__stack__.pop()

                if name == "default":
                    cls.error(
                        "You are not allowed to override the 'default' interface",
                        code=None,
                    )
                    return None
            else:
                cls.__stack__.clear()

                cls.error(
                    f"Rejected multiple interfaces declared in '{file_path.name}'",
                    code=None,
                )
                return None

        except Exception:
            stderr[0] << f"There are errors in '{file_path.name}':"
            stderr[0] << f"\n{traceback.format_exc(limit=-1)}\n"
            return None
        else:
            code_path = file_path.with_suffix(".pyc")

            with code_path.open(mode="wb") as file:
                marshal.dump(code, file)

            # Allow File Execution
            code_path.chmod(0o755)

            cls.log(f"Built '{file_path.name}'")

            return (name, code_path.name)

    @classmethod
    def remove(cls, *names: str, yes: bool = False):
        """"""
        cls._load()

        question = (
            f"This will remove {', '.join(names)}. Are you sure? (Y/n) ",
            {"Y": True, "n": False}
        )

        if yes or prompt(*question):
            for name in names:
                cls._remove(name)

        exit(EXIT_SUCCESS)    

    @classmethod
    def _remove(cls, name: str):
        """"""
        if name not in cls.options():
            cls.error(f"Unable to remove unknown interface {name!r}", code=None)
            return

        pack_path: Path = package_path()
        data_path: Path = pack_path.joinpath(".sat-api")
        code_path: Path = pack_path.joinpath(cls.__satref__[name])

        del cls.__satref__[name]

        with data_path.open(mode="w", encoding="utf8") as file:
            json.dump(cls.__satref__, file)

        code_path.unlink()


    @classmethod
    def clear(cls, yes: bool = False):
        """"""
        cls.remove(*cls.options(), yes=yes)

    @classmethod
    def options(cls) -> set[str]:
        cls._load()
        return set(cls.__satref__)

    @classmethod
    def complete(cls, answer: tuple[dict, float] | object) -> bool:
        """"""

        if isinstance(answer, tuple):
            if len(answer) == 2:
                if isinstance(answer[0], dict):
                    for k, v in answer[0].items():
                        if not isinstance(k, str):
                            return False
                        if not isinstance(v, int) or v not in {0, 1}:
                            return False
                else:
                    return False

                if isinstance(answer[1], float):
                    return True
                else:
                    return False
            else:
                return False
        else:
            return False

    @abstractmethod
    def solve(self, energy: Posiform, **params: dict) -> tuple[dict, float] | object:
        pass

    @staticmethod
    def error(message: str, code: int | None = EXIT_FAILURE):
        if stderr[0]:
            stderr[0] << f"API Error: {message}"
        if code is not None:
            exit(code)

    @staticmethod
    def warn(message: str, level: int = 1):
        if stdwar[level]:
            stdwar[level] << f"Warning: {message}"

    @staticmethod
    def log(message: str, level: int = 3):
        stdlog[level] << message


# -*- Default Interface -*-
class default(SatAPI):
    """"""

    ext = "json"

    def solve(self, energy: Posiform, indent: int | None = 4, **params: dict) -> str:
        """"""
        self.check_params(indent=indent)
        
        return energy.toJSON(indent=indent)

    def check_params(self, **params):
        """"""
        if "indent" in params:
            if params["indent"] is None:
                return
            elif not isinstance(params["indent"], int) or (params["indent"] < 0):
                self.error("Parameter 'indent' must be a non-negative integer")



"""
"""
# Future Imports
from __future__ import annotations

# Standard Library
import json

# -*- Satyrus -*-
from satyrus import SatAPI, Posiform


class qubo(SatAPI):
    """"""

    ext: str = "qubo.json"

    def solve(self, energy: Posiform, indent: int | None = None, **params: dict) -> str:
        """"""
        x, Q, c = energy.qubo()

        self.check_params(indent=indent)

        return json.dumps({"x": x, "Q": [list(q) for q in Q], "c": c}, indent=indent)

    def check_params(self, **params):
        """"""
        if "indent" in params:
            if params["indent"] is None:
                return
            elif not isinstance(params["indent"], int) or (params["indent"] < 0):
                self.error("Parameter 'indent' must be a non-negative integer")


from .base import SatType
from ..satlib import Source

class PythonObject(SatType):
    """"""

    def __init__(self, obj: object, *, source: Source = None, lexpos: int = None):
        SatType.__init__(self, source=source, lexpos=lexpos)
        self.obj = obj

__all__ = ['PythonObject']


""" All tokens are defined here.
    > sys.intern avoids double storage of a single string on memory. This also
      allows O(1) comparison using ´x is y´ syntax instead of O(k) ´x == y´.
"""
from sys import intern

## Ancillary variable token
T_AUX = intern("$")

## Extra Indexing variable token
T_ACC = intern(":")

## Logical
T_AND = intern("&")
T_OR = intern("|")
T_XOR = intern("^")
T_NXR = intern("!")
T_NOT = intern("~")

T_IFF = intern("<->")
T_IMP = intern("->")
T_RIMP = intern("<-")

## Quantifiers
T_FORALL = intern("@")
T_EXISTS = intern("$")
T_UNIQUE = intern("$!")

T_LOGICAL = {
    T_AND,
    T_OR,
    T_XOR,
    T_NOT,
    T_IFF,
    T_IMP,
    T_RIMP,
    ## Quantifiers
    T_FORALL,
    T_EXISTS,
    T_UNIQUE,
}


## Arithmetic
T_ADD = intern("+")
T_NEG = intern("-")
T_SUB = intern("-")
T_MUL = intern("*")
T_DIV = intern("/")
T_MOD = intern("%")
T_POW = intern("**")

T_ARITHMETIC = {T_ADD, T_NEG, T_MUL, T_DIV, T_MOD, T_POW}

## Indexing
T_IDX = intern("[]")

T_EXTRA = {T_IDX}

## Comparison
T_GT = intern(">")
T_LT = intern("<")
T_GE = intern(">=")
T_LE = intern("<=")
T_NE = intern("!=")
T_EQ = intern("==")

T_DICT = {
    "AND": T_AND,
    "RAND": T_AND,
    "OR": T_OR,
    "ROR": T_OR,
    "XOR": T_XOR,
    "RXOR": T_XOR,
    "NOT": T_NOT,
    "IFF": T_IFF,
    "IMP": T_IMP,
    "RIMP": T_RIMP,
    "ADD": T_ADD,
    "RADD": T_ADD,
    "POS": T_ADD,
    "SUB": T_NEG,
    "RSUB": T_NEG,
    "NEG": T_NEG,
    "MUL": T_MUL,
    "RMUL": T_MUL,
    "DIV": T_DIV,
    "RDIV": T_DIV,
    "TRUEDIV": T_DIV,
    "RTRUEDIV": T_DIV,
    "MOD": T_MOD,
    "RMOD": T_MOD,
    "GT": T_GT,
    "LT": T_LT,
    "GE": T_GE,
    "LE": T_LE,
    "NE": T_NE,
    "EQ": T_EQ,
    "IDX": T_IDX,
}

## Statements
RUN_INIT = intern("RUN_INIT")
CMD_INIT = (RUN_INIT,)
SYS_CONFIG = intern("SYS_CONFIG")
DEF_CONSTANT = intern("DEF_CONSTANT")
DEF_ARRAY = intern("DEF_ARRAY")
DEF_CONSTRAINT = intern("DEF_CONSTRAINT")
CMD_PYTHON = intern("CMD_PYTHON")
RUN_SCRIPT = intern("RUN_SCRIPT")
CMD_SCRIPT = (RUN_SCRIPT,)

## Sys config options
## To be used in `compiler.env` dict as keys.
DIR = intern("dir")
OUT = intern("out")
OPT = intern("opt")
PREC = intern("prec")
LOAD = intern("load")
EXIT = intern("exit")
ALPHA = intern("alpha")
EPSILON = intern("epsilon")
MAPPING = intern("mapping")
INDEXER = intern("indexer")

## Constraint Types
CONS_INT = intern("int")
CONS_OPT = intern("opt")



__SATYRUS_HELP = {
    "answer": "Initial guess/answer",
    "debug": "Enters debug mode (Sets verbosity to maximum level)",
    "source": "Source code file path",
    "legacy": "Intended to compile SATish code in legacy syntax",
    "solver": "Selects solver interface. For more information, refer to the 'sat-api' command",
    "output": "Output file destination",
    "verbose": "Compiler output verbosity, from 0 (default, no output, except errors and results) to 3 (complete compiler log)",
    "version": "Displays compiler version",
    "report": "Shows detailed sectioned elapsed time",
    "params": "Path to JSON file containing parameters for passing to Solver API",
    "clear": "Clears compiler cache",
}

__SAT_API_HELP = {
    "yes": "Skips user verification, answering yes to every question",
    "add": "Selects Python files (.py) where SatAPI interfaces are declared",
    "build": "Compiles the Python files (.py) into (.pyc) files",
    "clear": "Clears all user-defined interfaces",
    "remove": "Selects interfaces for removal",
}

def satyrus_help(key: str, **kwargs) -> str:
    if key in __SATYRUS_HELP:
        return __SATYRUS_HELP[key].format(**kwargs)
    return ""

def sat_api_help(key: str, **kwargs) -> str:
    if key in __SAT_API_HELP:
        return __SAT_API_HELP[key].format(**kwargs)
    return ""

__all__ = ["satyrus_help", "sat_api_help"]